"""
New version of Reason, using a TraitsUI-based approach

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.mods import *
#pf = EnzoStaticOutput("/Users/matthewturk/Research/data/galaxy1200.dir/galaxy1200")

from enthought.traits.api import \
    HasTraits, List, Instance, Str, Float, Any, Code, PythonValue, Int, CArray, \
    Property, Enum, cached_property, DelegatesTo, Callable, Array, \
    Button
from enthought.traits.ui.api import \
    Group, VGroup, HGroup, Tabbed, View, Item, ShellEditor, InstanceEditor, ListStrEditor, \
    ListEditor, VSplit, VFlow, HSplit, VFold, ValueEditor, TreeEditor, TreeNode, RangeEditor, \
    EnumEditor, Handler, Controller, DNDEditor
from enthought.traits.ui.menu import \
    Menu, Action, Separator, OKCancelButtons, OKButton
from enthought.pyface.action.api import \
    ActionController
from enthought.tvtk.pyface.scene_editor import SceneEditor
from enthought.tvtk.pyface.api import \
    DecoratedScene
from enthought.tvtk.pyface.scene_model import SceneModel
from enthought.traits.ui.wx.range_editor import SimpleSliderEditor

from plot_editors import Figure, MPLFigureEditor, MPLVMPlotEditor, Axes

from yt.raven.PlotTypes import VMPlot, ProjectionPlot, SlicePlot

import traceback
from tvtk_interface import \
    HierarchyImporter, YTScene

class PlotCreationHandler(Controller):
    main_window = Instance(HasTraits)
    pnode = Instance(HasTraits)

    format = Str
    plot_type = Any
    
    def close(self, info, is_ok):
        if not is_ok:
            super(Controller, self).close(info, True)
            return
        spt = self.plot_type(plot_spec=self.model, pf=self.pnode.pf,
                           name=self.format % (self.model.axis))
        self.pnode.data_objects.append(spt)
        self.main_window.plot_frame_tabs.append(spt)
        spt.plot

class VTKSceneCreationHandler(PlotCreationHandler):
    importer = Instance(HierarchyImporter)

    def close(self, info, is_ok):
        if is_ok: 
            yt_scene = YTScene(importer=self.importer,
                scene=SceneModel())
            spt = VTKDataObject(name = "VTK: %s" % self.pnode.pf,
                    scene=yt_scene.scene,
                    yt_scene=yt_scene)
            self.pnode.data_objects.append(spt)
            self.main_window.plot_frame_tabs.append(spt)
        super(Controller, self).close(info, True)
        return True


class DataObject(HasTraits):
    name = Str

class VTKDataObject(DataObject):
    yt_scene = Instance(YTScene)
    scene = DelegatesTo("yt_scene")
    add_contours = Button
    add_isocontour = Button
    add_x_plane = Button
    add_y_plane = Button
    add_z_plane = Button
    edit_camera = Button
    edit_operators = Button
    edit_pipeline = Button
    center_on_max = Button
    operators = DelegatesTo("yt_scene")
    traits_view = View(
            Item("scene", editor = 
        SceneEditor(scene_class=DecoratedScene),
                    resizable=True, show_label=False),
            HGroup(Item("add_contours", show_label=False),
                   Item("add_isocontour", show_label=False),
                   Item("add_x_plane", show_label=False),
                   Item("add_y_plane", show_label=False),
                   Item("add_z_plane", show_label=False),
                   Item("edit_camera", show_label=False),
                   Item("edit_operators", show_label=False),
                   Item("edit_pipeline", show_label=False),
                   Item("center_on_max", show_label=False),
                ),
            )

    operators_edit = View(
        Item("operators", style='custom', show_label=False,
             editor=ListEditor(editor=InstanceEditor(),
                               use_notebook=True),
              name="Edit Operators"),
        height=500.0, width=500.0, resizable=True)
    
    def _edit_camera_fired(self):
        self.yt_scene.camera_path.edit_traits()

    def _edit_operators_fired(self):
        self.edit_traits(view='operators_edit')

    def _edit_pipeline_fired(self):
        from enthought.tvtk.pipeline.browser import PipelineBrowser
        pb = PipelineBrowser(self.scene)
        pb.show()

    def _add_contours_fired(self):
        self.yt_scene.add_contour()

    def _add_isocontour_fired(self):
        self.yt_scene.add_isocontour()

    def _add_x_plane_fired(self):
        self.yt_scene.add_x_plane()

    def _add_y_plane_fired(self):
        self.yt_scene.add_y_plane()

    def _add_z_plane_fired(self):
        self.yt_scene.add_z_plane()

    def _center_on_max_fired(self):
        self.yt_scene.do_center_on_max()

class ParameterFile(HasTraits):
    pf = Instance(EnzoStaticOutput)
    data_objects = List(Instance(DataObject))
    name = Str

    def _name_default(self):
        return str(self.pf)

    def do_slice(self):
        cons_view = View(
                Item('axis'), 
                Item('center'), 
                Item('field', editor=EnumEditor(name='field_list')),
                buttons=OKCancelButtons, title="Slicer: %s" % self.pf)
        ps = SlicePlotSpec(pf=self.pf)
        hand = PlotCreationHandler(main_window=mw, pnode=self, model=ps,
                                   plot_type=SlicePlotTab, format="Slice: %s")
        ps.edit_traits(cons_view, handler=hand)

    def do_proj(self):
        cons_view = View(
                Item('axis'), 
                Item('field', editor=EnumEditor(name='field_list')),
                Item('weight_field', editor=EnumEditor(name='none_field_list')),
                buttons=OKCancelButtons, title="Projector: %s" % self.pf)
        ps = ProjPlotSpec(pf=self.pf)
        hand = PlotCreationHandler(main_window=mw, pnode=self, model=ps,
                                   plot_type=ProjPlotTab, format="Proj: %s")
        ps.edit_traits(cons_view, handler=hand)

    def do_vtk(self):
        from tvtk_interface import HierarchyImporter, \
            HierarchyImportHandler
        importer = HierarchyImporter(pf=self.pf, max_level=self.pf.h.max_level)
        importer.edit_traits(handler = VTKSceneCreationHandler(
            main_window=mw, pnode=self, importer = importer))

class ParameterFileCollection(HasTraits):
    parameter_files = List(Instance(ParameterFile))
    name = Str
    collection = Any

    def _parameter_files_default(self):
        my_list = []
        for f in self.collection:
            try:
                pf = EnzoStaticOutput(f)
                my_list.append(
                    ParameterFile(pf=pf, 
                            data_objects = []))
            except IOError: pass
        return my_list

    def _name_default(self):
        return str(self.collection)

class ParameterFileCollectionList(HasTraits):
    parameter_file_collections = List(Instance(ParameterFileCollection))

    def _parameter_file_collections_default(self):
        return [ParameterFileCollection(collection=c)
                for c in fido.GrabCollections()]

class DataObjectList(HasTraits):
    data_objects = List(Str)

    traits_view = View(
              Item('data_objects', show_label=False,
                   editor=ListStrEditor())
               )

    def _data_objects_default(self):
        return ['a','b','c']

class PlotFrameTab(DataObject):
    figure = Instance(Figure)

class VMPlotSpec(HasTraits):
    pf = Instance(EnzoStaticOutput)
    field = Str('Density')
    field_list = Property(depends_on = 'pf')

    center = Array(shape=(3,), dtype='float64')
    axis = Enum(0,1,2)

    @cached_property
    def _get_field_list(self):
        fl = self.pf.h.field_list
        df = self.pf.h.derived_field_list
        fl.sort(); df.sort()
        return fl + df

    def _center_default(self):
        return self.pf.h.find_max("Density")[1]

class SlicePlotSpec(VMPlotSpec):
    pass

class ProjPlotSpec(VMPlotSpec):
    weight_field = Str("None")
    none_field_list = Property(depends_on = 'field_list')

    @cached_property
    def _get_none_field_list(self):
        return ["None"] + self.field_list

class VMPlotTab(PlotFrameTab):
    pf = Instance(EnzoStaticOutput)
    figure = Instance(Figure, args=())
    field = DelegatesTo('plot_spec')
    field_list = DelegatesTo('plot_spec')
    plot = Instance(VMPlot)
    axes = Instance(Axes)
    disp_width = Float(1.0)
    unit = Str('unitary')
    min_width = Property(Float, depends_on=['pf','unit'])
    max_width = Property(Float, depends_on=['pf','unit'])
    unit_list = Property(depends_on = 'pf')
    smallest_dx = Property(depends_on = 'pf')

    traits_view = View(VGroup(
            HGroup(Item('figure', editor=MPLVMPlotEditor(),
                     show_label=False)),
            HGroup(Item('disp_width',
                     editor=RangeEditor(format="%0.2e",
                        low_name='min_width', high_name='max_width',
                        mode='logslider', enter_set=True),
                     show_label=False, width=400.0),
                   Item('unit',
                      editor=EnumEditor(name='unit_list')),),
            HGroup(Item('field',
                      editor=EnumEditor(name='field_list')),
                )),
             resizable=True)

    def __init__(self, **traits):
        super(VMPlotTab, self).__init__(**traits)
        self.axes = self.figure.add_subplot(111, aspect='equal')

    def _field_changed(self, old, new):
        self.plot.switch_z(new)
        self._redraw()

    @cached_property
    def _get_min_width(self):
        return 50.0*self.smallest_dx*self.pf[self.unit]

    @cached_property
    def _get_max_width(self):
        return self.pf['unitary']*self.pf[self.unit]

    @cached_property
    def _get_smallest_dx(self):
        return self.pf.h.get_smallest_dx()

    @cached_property
    def _get_unit_list(self):
        return self.pf.units.keys()

    def _unit_changed(self, old, new):
        self.disp_width = self.disp_width * self.pf[new]/self.pf[old]

    def _disp_width_changed(self, old, new):
        self.plot.set_width(new, self.unit)
        self._redraw()

    def _redraw(self):
        self.figure.canvas.draw()

    def recenter(self, event):
        xp, yp = event.xdata, event.ydata
        dx = abs(self.plot.xlim[0] - self.plot.xlim[1])/self.plot.pix[0]
        dy = abs(self.plot.ylim[0] - self.plot.ylim[1])/self.plot.pix[1]
        x = (dx * xp) + self.plot.xlim[0]
        y = (dy * yp) + self.plot.ylim[0]
        xi = lagos.x_dict[self.axis]
        yi = lagos.y_dict[self.axis]
        cc = self.center[:]
        cc[xi] = x; cc[yi] = y
        self.plot.data.center = cc[:]
        self.plot.data.set_field_parameter('center', cc.copy())
        self.center = cc

class SlicePlotTab(VMPlotTab):
    plot_spec = Instance(SlicePlotSpec)

    axis = DelegatesTo('plot_spec')
    center = DelegatesTo('plot_spec')
    
    plot = Instance(SlicePlot)

    def _plot_default(self):
        coord = self.center[self.axis]
        sl = self.pf.h.slice(self.axis, coord, center=self.center[:])
        sp = SlicePlot(sl, self.field, self.figure, self.axes)
        self.figure.canvas.draw()
        return sp

    def _center_changed(self, old, new):
        #traceback.print_stack()
        if na.all(na.abs(old - new) == 0.0): return
        print na.abs(old-new)
        print "Re-slicing", old, new
        pp = self.center
        self.plot.data.reslice(pp[self.axis])
        self.plot._refresh_display_width()
        self.figure.canvas.draw()

class ProjPlotTab(VMPlotTab):
    plot_spec = Instance(ProjPlotSpec)

    axis = DelegatesTo('plot_spec')
    center = DelegatesTo('plot_spec')
    weight_field = DelegatesTo('plot_spec')

    plot = Instance(ProjectionPlot)

    def _plot_default(self):
        self.field = self.field[:]
        self.weight_field = self.weight_field[:]
        wf = self.weight_field
        if str(wf) == "None": wf = None
        proj = self.pf.h.proj(self.axis, self.field, wf,
                        center=self.center[:])
        pp = ProjectionPlot(proj, self.field, self.figure, self.axes)
        self.figure.canvas.draw()
        return pp

    def _center_changed(self, old, new):
        self.plot._refresh_display_width()

class SphereWrapper(DataObject):
    radius = Float
    unit = Str

class MainWindow(HasTraits):
    parameter_file_collections = Instance(ParameterFileCollectionList)
    parameter_files = Instance(ParameterFileCollection)
    plot_frame_tabs = List(Instance(DataObject))
    open_parameterfile = Button
    shell = PythonValue

    def _shell_default(self):
        return globals()
    notebook_editor = ListEditor(editor=InstanceEditor(editable=True),
                                 use_notebook=True)

    traits_view = View(VSplit(
                    HSplit(VGroup(
                       Item('parameter_file_collections', 
                            width=120.0, height=500.0,
                            show_label=False,
                            editor = TreeEditor(editable=False,
                    nodes=[
                        TreeNode(node_for=[ParameterFileCollectionList],
                                 children='parameter_file_collections',
                                 label="=Data Collections"),
                        TreeNode(node_for=[ParameterFileCollection],
                                 children='parameter_files',
                                 label="name",
                                 view=View()),
                        TreeNode(node_for=[ParameterFile],
                                 children='data_objects',
                                 label="name",
                                 menu = Menu(Action(name='Slice',
                                                    action='object.do_slice'),
                                             Action(name='Project',
                                                    action='object.do_proj'),
                                             Action(name='VTK',
                                                    action='object.do_vtk')),
                                 view=View()),
                        TreeNode(node_for=[DataObject],
                                 children='',
                                 label="name"),
                                ], show_icons=False),),
                        Item('open_parameterfile', show_label=False)),
                       Item('plot_frame_tabs', style='custom',
                            editor = notebook_editor,
                            show_label=False, height=500.0, width=500.0),
                    ),
                    HGroup(
                       Item('shell', editor=ShellEditor(share=True),
                            show_label=False, height=120.0),
                    ),
                ),
               resizable=True, width=800.0, height=660.0,
               title="reason v2 [prototype]")

    def _open_parameterfile_fired(self):
        print "OPENING"

    def _parameter_file_collections_default(self):
        return ParameterFileCollectionList()

class YTScript(HasTraits):
    code = Code
    traits_view = View(Item('code', show_label=False),
                       height=0.8, width=0.8, resizable=True,
                       buttons=OKCancelButtons)

class ObjectViewer(HasTraits):
    to_view=Any
    traits_view = View(
            Item('to_view', editor=ValueEditor(), show_label=False),
                     resizable=True, height=0.8, width=0.8)

def view_object(obj):
    ObjectViewer(to_view=obj).edit_traits()

def run_script():
    my_script = YTScript()
    my_script.edit_traits()
    return my_script

class event_mock(object):
    inaxes = True
    button = 3

dol = DataObjectList()
mw = MainWindow(plot_frame_tabs = [])
mw.configure_traits()
#mw.edit_traits()



"""
Raven
=====

Raven is the plotting interface, with support for several
different engines.  Well, two for now, but maybe more later.
Who knows?

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from yt.config import ytcfg
from yt.logger import ravenLogger as mylog
from yt.arraytypes import *
from yt.funcs import *
import yt.lagos as lagos

import matplotlib
matplotlib.rc('contour', negative_linestyle='solid')

import matplotlib.image
import matplotlib.ticker
import matplotlib.axes
import matplotlib.figure
import matplotlib._image
import matplotlib.colors
import matplotlib.colorbar
import matplotlib.cm
import matplotlib.collections

import time, types, string, os

# @todo: Get rid of these
axis_labels = [('y','z'),('x','z'),('x','y')]
axis_names = {0: 'x', 1: 'y', 2: 'z'}

vm_axis_names = {0:'x', 1:'y', 2:'z', 3:'dx', 4:'dy'}

from ColorMaps import raven_colormaps, add_cmap, check_color

import PlotTypes
be = PlotTypes

from Callbacks import *
from FixedResolution import *

if 'cmapnames' in dir(matplotlib.cm):
    color_maps = matplotlib.cm.cmapnames + raven_colormaps.keys()
else:
    color_maps = matplotlib.cm._cmapnames + raven_colormaps.keys()
default_cmap = ytcfg.get("raven", "colormap")
if default_cmap != "jet":
    mylog.debug("Setting default colormap to %s", default_cmap)
    matplotlib.rc('image', cmap=default_cmap)

from plot_collection import *



"""
Python kD Tree

Author: Michael Knight
Affiliation: ?
Note: Code is based on <http://sites.google.com/site/mikescoderama/Home/kd-tree-knn>,
with periodicity added and a few other cosmetic changes. There is no contact
infomation on that page, and this code's license is a bit uncertain.
Author: Stephen Skory <stephenskory@yahoo.com>
Affiliation: UCSD Physics/CASS
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.lagos import *

from bisect import insort

class Point:pass

class Node:

    def _printOut(self,depth = 0):
        """
        iteratively print out the kd tree nodes, depth really isn't a user-set
        parameter.
        """
        for i in range(0,depth):
            print "....",
        print "point count =", self.pointCount, "rect =", \
            self.hyperRect._nodeToString(), '\n' #"points =",self.points
        if (self.leftChild is not None):
            for i in range(0,depth):
                print "....",
            print "left: "
            self.leftChild._printOut(depth+1)
        if (self.rightChild is not None):
            for i in range(0,depth):
                print "....",
            print "right: "
            self.rightChild._printOut(depth+1)
    
    def _buildBoundingHyperRect(self,points):
        self.hyperRect = HyperRect()
        self.hyperRect._buildBoundingHyperRect(points)
 
def getFastDistance(a,b,period):
    """
    returns the square of the distance between points a and b,
    using periodic boundary conditions 'period'.
    """
    dim = len(b);
    total = 0;
    for i in range(0,dim):
        delta = min(abs(a[i] - b[i]), period[i] - abs(a[i] - b[i]));
        total = total + (delta *delta)
    return total
 
class Neighbors:
    
    def _addNeighbors(self,node,query,period):
        """
        for each point in this node, calculate the distance to the query point,
        if it's close enough add it to the list .points using insort
        """
        for i in range(0,node.pointCount):
            dist = getFastDistance(node.points[i].data,query,period)
            if (dist < self.minDistanceSquared):
                item = [dist,node.points[i]]
                insort(self.points,item)
                if (len(self.points) > self.k):
                    self.points = self.points[0:self.k]
 
        if (len(self.points) == self.k):
            self.minDistanceSquared = self.points[self.k-1][0]
        return;
 
class HyperRect:
    def _buildBoundingHyperRect(self,points):
        """
        find the extremities of the hypercube.
        """
        self.k = len(points[0].data)
        self.dims  = range(0,self.k) 
        high = points[0].data[:]
        low = points[0].data[:]
        for i in range(0,len(points)):
            for j in self.dims:
                point = points[i].data[j]
                if (high[j]  < point):
                    high[j] = point
                if (low[j] > point):
                    low[j] = point
        self.high = high
        self.low = low
        return 
 
    def _getWidestDimension(self):
        """
        get the widest dimension of the points in order to find the dimension
        to bisect.
        """
        widest =0
        widestDim =-1
        for i in self.dims:
            width = self.high[i] -  self.low[i]
            if (width > widest):
                widestDim =i
                widest = width
        self.widest = widest;
        self.widestDim = widestDim;    
        return self.widestDim
    
    def _getWidestDimensionWidth(self): # I don't know why this is here.
        return self.widest
    
    def _nodeToString(self):
        return "high =",self.high,"low =",self.low,
    
    def _getMinDistance(self,query,period):
        """
        find the minimum distance squared to a corner of the hypercube
        from the query point using periodicity.
        """
        total = 0.0
        for i in self.dims:
            delta = 0.0
            min_high = min(abs(query[i] - self.high[i]), \
                period[i] - abs(query[i] - self.high[i]))
            min_low = min(abs(query[i] - self.low[i]), \
                period[i] - abs(query[i] - self.low[i]))
            delta = min(min_low,min_high)
            total = total + (delta*delta)
        return total;

def buildKdHyperRectTree(points,rootMin=3):
    """
    Recursively build the kdTree, adding nodes as needed until all have no more
    than rootMin points. The final nodes are called leafs, which contain the
    point data.
    """
    if (points is None or len(points) ==0):
        return None
    n = Node() # make a new node
    n._buildBoundingHyperRect(points)  # build the hyper rect for these points
                        # this will find the top left and botom
                        # right of all given points.

    leaf = False
    # If the size of points is small enough, this node is a leaf
    if len(points) <= rootMin:
        leaf = True
    splitDim  = -1
 
    if (not leaf):
        # get the widest dimension to split n to maximize splitting affect
        splitDim = n.hyperRect._getWidestDimension()
        # do we have a bunch of children at the same point?
        if (n.hyperRect._getWidestDimensionWidth() == 0.0):
            left = True 
    #init the node
    n.pointCount = len(points)
    n.points = None
    n.leftChild = None
    n.rightChild = None
    n.points = None
 
    if (leaf or len(points)==0):
        n.points = points # we are a leaf so just store all points in the rect
    else:
        # sort by the best split dimension
        temp = []
        for index,p in enumerate(points):
            insort(temp,[p.data[splitDim],index])
        temp2 = points[:]
        for index,t in enumerate(temp):
            temp2[index] = points[t[1]]
        points = temp2[:]
        del temp, temp2
        #points.sort(key=lambda points: points.data[splitDim])
        median = len(points)/2     # get the median
        # and split left for smaller values in splitDim, right for larger
        n.leftChild = buildKdHyperRectTree(points[0:(median+1)], rootMin)
        if (median+1 < len(points)):
            n.rightChild = buildKdHyperRectTree(points[median+1:], rootMin)
    return n;
 
def getKNN(query,node, neighbors,distanceSquared,period):
    """
    Recursively walk the kd tree, limited by *distanceSquared* to the extrema of
    the hypercubes, only finding distances to the query point in leaf nodes.
    *neighbors* is a Neighbors object, and needs to be initialized bofore
    calling this. *period* is a list or array of the period for each dimension
    of the hypercube.
    """
    
    # test to see if the query point is inside this node
    # <= and >= on both ends is okay, it's better to be inclusive and it
    # prevents problems with particles on boundaries
    for i in node.hyperRect.dims:
        if query[i] <= node.hyperRect.high[i] and \
                query[i] >= node.hyperRect.low[i]:
            inside = True
        else:
            inside = False
            break
    
    # if this node is close enough (the distances are calculated in the previous
    # iteration), or if the query point is inside the node, continue on
    if True or inside:
        # leafs don't have children, so this tests to see if this node
        # is a leaf, and if it is a leaf, calculate distances to the query point
        if (node.leftChild is None):
            # add to neighbors.points
            neighbors._addNeighbors(node,query,period)
        # if this node is not a leaf, find out the distance to its children,
        # and then continue the iteration down the kd tree.
        else:
            distLeft = node.leftChild.hyperRect._getMinDistance(query,period)
            distRight = node.rightChild.hyperRect._getMinDistance(query,period)
            if (distLeft < distRight):
                getKNN(query,node.leftChild,neighbors,distLeft,period)
                getKNN(query,node.rightChild,neighbors,distRight,period)
            else:
                getKNN(query,node.rightChild,neighbors,distRight,period)
                getKNN(query,node.leftChild,neighbors,distLeft,period)



class AnalysisTask(object):

    def __init__(self, *args, **kwargs):
        # This should only get called if the subclassed object
        # does not override
        if len(args) + len(kwargs) != len(self._params):
            raise RuntimeError
        self.__dict__.update(zip(self._params, args))
        self.__dict__.update(kwargs)

    def __repr__(self):
        # Stolen from AMRData.__repr__
        s = "%s: " % (self.__class__.__name__, self._analysis_type)
        s += ", ".join(["%s=%s" % (i, getattr(self,i))
                       for i in self._params])
        return s

class MaximumValue(AnalysisTask):
    _params = ['field']

    def eval(self, data_object):
        v = data_object.quantities["MaxLocation"](
                self.field, lazy_reader=True)[0]
        return v

class ParameterValue(AnalysisTask):
    _params = ['parameter']

    def __init__(self, parameter, cast=None):
        self.parameter = parameter
        if cast is None:
            cast = lambda a: a
        self.cast = cast

    def eval(self, pf):
        return self.cast(pf.get_parameter(self.parameter))

class CurrentTimeYears(AnalysisTask):
    _params = []

    def eval(self, pf):
        return pf["InitialTime"] * pf["years"]

class SliceDataset(AnalysisTask):
    _params = ['field', 'axis']

    def eval(self, pf):
        pass

class SlicePlotDataset(AnalysisTask):
    _params = ['field', 'axis', 'center']

    def __init__(self, *args, **kwargs):
        import yt.raven
        self.raven = yt.raven
        AnalysisTask.__init__(self, *args, **kwargs)

    def eval(self, pf):
        pc = self.raven.PlotCollection(pf, center = self.center)
        pc.add_slice(self.field, self.axis)
        return pc.save()[0]



"""
Helpful recipes to make tasty AMR desserts

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# Named imports
import yt.lagos as lagos
import yt.raven as raven
from yt.funcs import *
import numpy as na
import os.path, inspect, types
from functools import wraps
from yt.logger import ytLogger as mylog

# These next couple functions are for 'fixing' arguments - accepting arguments
# of a couple different types and styles, and trying to coerce them into what
# we want.

def _fix_pf(pf):
    if isinstance(pf, lagos.StaticOutput): return pf
    if os.path.exists("%s.hierarchy" % pf):
        return lagos.EnzoStaticOutput(pf)
    elif os.path.isdir("%s" % pf) and \
         os.path.exists("%s/%s" % (pf,pf)):
        return lagos.EnzoStaticOutput("%s/%s" % (pf,pf))
    elif os.path.isdir("%s.dir" % pf) and \
         os.path.exists("%s.dir/%s" % (pf,pf)):
        return lagos.EnzoStaticOutput("%s.dir/%s" % (pf,pf))
    elif pf.endswith(".hierarchy"):
        return lagos.EnzoStaticOutput(pf[:-10])
    elif os.path.isdir("%s" % pf) and \
         os.path.exists("%s/Header" % pf):
        return lagos.OrionStaticOutput(pf)
    else:
        raise IOError(pf)

__pf_centers = {}
def _fix_center(pf, center):
    if center is not None and iterable(center):
        center = na.array(center)
    else:
        if pf['CurrentTimeIdentifier'] in __pf_centers:
            center = __pf_centers[pf['CurrentTimeIdentifier']]
        else:
            center = pf.h.find_max("Density")[1]
            __pf_centers[pf['CurrentTimeIdentifier']] = center
    return center

def _fix_radius(pf, radius):
    if radius is not None:
        if iterable(radius):
            return radius[0] / pf[radius[1]]
        return radius
    mylog.info("Setting radius to be 0.1 of the box size")
    # yt-generalization : needs to be changed to 'unitary'
    return 0.1 / pf["1"]

def _fix_width(pf, width):
    if width is not None:
        if iterable(width):
            return width[0] / pf[width[1]]
        return width
    mylog.info("Setting width to be the full box")
    # yt-generalization : needs to be changed to 'unitary'
    return 1.0 / pf["1"]

def _fix_axis(pf, axis):
    if axis is None:
        raise ValueError("You need to specify an Axis!")
    elif isinstance(axis, types.IntType) and (0 <= x <= 2):
        return axis
    elif isinstance(axis, types.StringTypes) and \
         axis.upper() in 'XYZ':
        return 'XYZ'.find(axis.upper())
    else:
        raise ValueError("Invalid Axis specified.")


_arg_fixer = {
                'center' : (True, _fix_center),
                'radius' : (True, _fix_radius),
                'width' : (True, _fix_width),
                'axis' : (True, _fix_axis),
             }

#
# * Need to change filename to be an argument in the descriptor
# * Need to add cmap to the descriptor
#

class FakePlotCollection(object):
    def __init__(self, fig):
        self.fig = fig
    def save(self, filename):
        self.fig.savefig(filename)

def fix_plot_args(plot_func):
    @wraps(plot_func)
    def call_func(self, pf, **kwargs):
        pf = _fix_pf(pf)
        fkwargs = inspect.getargspec(plot_func)[0]
        for arg in fkwargs:
            if arg in _arg_fixer:
                needs, fixer = _arg_fixer[arg]
                if arg in kwargs:
                    kwargs[arg] = fixer(pf, kwargs[arg])
                elif needs:
                    kwargs[arg] = fixer(pf, None)
        retval = plot_func(self, pf, **kwargs)
        if 'filename' in kwargs and 'filename' in fkwargs:
            retval.save(kwargs['filename'])
        return retval
    return call_func

which_pc = raven.PlotCollection

# to add:
#   zoom movies

class _RecipeBook(object):

    @fix_plot_args
    def mass_enclosed_radius(self, pf, center=None, radius=None, radius_unit="Radiuspc",
                             filename = None):
        pc = which_pc(pf, center=center)
        p = pc.add_profile_sphere(radius, '1', 
                                  [radius_unit, "CellMassMsun"],
                                  accumulation=True, weight=None)
        return pc

    @fix_plot_args
    def mass_enclosed_field(self, pf, field="Density", center=None, radius=None,
                            weight="CellMassMsun", filename = None):
        pc = which_pc(pf, center=center)
        p = pc.add_profile_sphere(radius, '1', 
                                  ["Radius", "CellMassMsun"],
                                  accumulation=True, weight=None)
        p.switch_z(field, weight=weight)
        p.switch_x("CellMassMsun")
        return pc

    @fix_plot_args
    def slice_axis(self, pf, field="Density", center=None, width=None, filename=None, axis=None, coord=None):
        pc = which_pc(pf, center=center)
        p = pc.add_slice(field, axis, coord=coord)
        pc.set_width(width, '1')
        return pc

    @fix_plot_args
    def slice(self, pf, field="Density", center=None, width=None, filename=None):
        pc = which_pc(pf, center=center)
        for axis in range(3): pc.add_slice(field, axis, coord=center[axis])
        pc.set_width(width, '1')
        return pc

    def _multiplot(self, pc, func, pf, field="Density", center=None, width=None, filename=None):
        import pylab
        pylab.subplots_adjust(wspace=0.0, hspace=0.0,
                              top=1.0, bottom=0.0,
                              left=0.0, right=1.0)
        pylab.gcf().set_size_inches((12,12))
        pylab.clf()
        mi, ma = 1e30, -1e30
        for axis in range(3):
            axes = pylab.subplot(2,2,axis+1)
            func(field, axis, coord=center[axis], axes=axes)
        pc.set_width(width, '1')
        for axis in range(3):
            mi = min(mi, pc.plots[axis].image._A.min())
            ma = max(ma, pc.plots[axis].image._A.max())
        pc.set_zlim(mi,ma)
        cba = pylab.axes([0.55, 0.05, 0.02, 0.40])
        cb = pylab.colorbar(cax=cba, mappable=pc.plots[-1].image)
        pc.plots[-1].colorbar = cb
        pc.plots[-1].autoset_label()
        return FakePlotCollection(pylab.gcf())

    @fix_plot_args
    def multiplot_slice(self, pf, field="Density", center=None, width=None, filename=None):
        pc = which_pc(pf, center=center)
        func = pc.add_slice
        return self._multiplot(pc, func, pf, field, center, width, filename)

    @fix_plot_args
    def multiplot_proj(self, pf, field="Density", center=None, width=None, filename=None):
        pc = which_pc(pf, center=center)
        func = pc.add_projection
        return self._multiplot(pc, func, pf, field, center, width, filename)

    def dispatch(self):
        pass

RecipeBook = _RecipeBook()

if __name__ == "__main__":
    RecipeBook.dispatch()



"""
Handling of sets of outputs.

We don't do any instantiation, or touching of lagos, etc.

Very simple nowadays.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.fido import *

from stat import ST_CTIME

class OutputCollection:
    def __init__(self, title):
        self.title = title
        self.output_names = na.array(())
        self.output_time_ids = na.array((), dtype='int64')
        self.output_times = na.array((), dtype='float64')

    def read_in(self, filename):
        self._last_read_filename = filename
        lines = open(filename).readlines()
        output_lines = [line for line in lines if line.startswith("Output:")]
        outputs = [line.split(":")[1] for line in output_lines]
        output_time_ids = [int(line.split(":")[2]) for line in output_lines]
        output_times = [float(line.split(":")[3]) for line in output_lines]

        self.output_names = na.array(outputs)
        self.output_time_ids = na.array(output_time_ids, dtype='int64')
        self.output_times = na.array(output_times, dtype='float64')

    def write_out(self):
        path=ytcfg.get("fido","rundir")
        if not os.path.isdir(path): os.makedirs(path)
        fn = os.path.join(path, "runF_%s" % (self.title))
        f = open(fn, "w")
        for i, output in enumerate(self.output_names):
            f.write("Output:%s:%s:%s\n" % ( \
                        os.path.abspath(output), \
                        self.output_time_ids[i], \
                        self.output_times[i]))
        f.close()

    def sort(self):
        order = na.argsort(self.output_times)
        self.output_names = self.output_names[order]
        self.output_times = self.output_times[order]
        self.output_time_ids = self.output_time_ids[order]

    def add_output(self, filename):
        # We're passing in *just* filenames here.  So, we simply snag the
        # two appropriate lines in the parameter file.
        if filename in self.output_names: return
        time = get_parameter_line(filename, "InitialTime").split()[-1]
        # Implement exception catching
        try:
            timeID = get_parameter_line(filename, "CurrentTimeIdentifier").split()[-1]
        except:
            timeID = int(os.stat(filename)[ST_CTIME])
        self.output_names = \
            na.array(self.output_names.tolist() + [filename])
        self.output_time_ids = \
            na.array(self.output_time_ids.tolist() + [int(timeID)], dtype='int64')
        self.output_times = \
            na.array(self.output_times.tolist() + [float(time)], dtype='float64')
        self.sort()

    def get_before(self, time):
        return na.where(self.output_times <= time)[0]

    def get_after(self, time):
        return na.where(self.output_times > time)[0]

    def __delitem__(self, key):
        if isinstance(key, types.StringType):
            id = self.index(key)
        elif isinstance(key, types.IntType):
            id = key
        self.output_names = na.array(self.output_names[:id].tolist() \
                                  + self.output_names[id+1:].tolist())
        self.output_times = na.array(self.output_times[:id].tolist() \
                                  + self.output_times[id+1:].tolist())
        self.output_time_ids = na.array(self.output_time_ids[:id].tolist() \
                                  + self.output_time_ids[id+1:].tolist())

    def __getitem__(self, key):
        """
        Based on the type of input, we either return based on index or
        basename.
        """
        if isinstance(key, types.StringType):
            index = self.index(key)
        elif isinstance(key, types.IntType):
            index = key
        # This fails, but I don't know how to fix it.
        a = (self.output_names[index],  \
             self.output_times[index],  \
             self.output_time_ids[index] )
        return a

    def index(self, key):
        t = os.path.basename(key)
        # Find out the index
        index = None
        for i in range(self.output_names.shape[0]):
            if os.path.basename(self.output_names[i]) \
                == os.path.basename(t):
                index = i
                break
        if index == None:
            raise KeyError
        return index

    def __repr__(self):
        return self.title

    def __len__(self):
        return len(self.output_names)

    def __iter__(self):
        for i in range(len(self)):
            yield self.output_names[i]

    def keys(self):
        return self.output_names.tolist()

    def __convert_args(self, args, kwargs, my_dict):
        new_args = list(args)[:]
        new_kwargs = {}
        for j, arg in enumerate(args):
            if isinstance(arg, types.StringType):
                new_args[j] = arg % myDict
        for key in kwargs.keys():
            if isinstance(kwargs[key], types.StringType):
                new_kwargs[key] = kwargs[key] % myDict
        return new_args, new_kwargs

    def run_functions(self, function, args = None, kwargs = None):
        if args is None: args = []
        if kwargs is None: kwargs = {}
        import yt.lagos as lagos # We import *here* so that we only import if we need it
        for i,o in enumerate(self):
            # Now we format string the various args and kwargs
            myDict = {'fn':str(o), 'index':i, 'time':self.output_times[i],
                      'timeID':self.output_time_ids[i]}
            newa, newk = self.__convert_args(args, kwargs)
            print args, kwargs
            function(*newa, **newk)

def GrabCollections(path=None):
    if not path: path=ytcfg.get("fido","rundir")
    if not os.path.isdir(path): os.makedirs(path)
    ocs = []
    for file in glob.glob(os.path.join(path,"runF_*")):
        title=os.path.basename(file)
        if title.startswith("runF_"): title = title[5:]
        ocs.append(OutputCollection(title))
        ocs[-1].read_in(file)
    return ocs



""" Original with Weave:
Doing contour 9 / 10 (4.56276e-14 9.91625e-14)
yt.lagos   INFO       2008-04-10 17:04:45,085 Getting field Density from 197
yt.lagos   INFO       2008-04-10 17:04:46,284 Contouring over 1061337 cells with 30115 candidates
yt.lagos   INFO       2008-04-10 17:05:17,880 Getting field tempContours from 197
yt.lagos   INFO       2008-04-10 17:05:18,305 Identified 5 contours between 4.56276e-14 and 2.15510e-13
yt.lagos   INFO       2008-04-10 17:05:18,312 Getting field Contours from 197
yt.lagos   INFO       2008-04-10 17:05:19,330 Getting field GridIndices from 197
        Contour id 0.0 has: 4.56297e-14 9.88512e-14  (2351) (4 grids, 11081.0 11087.0)
        Contour id 1.0 has: 4.58381e-14 5.41707e-14  (13) (1 grids, 11079.0 11079.0)
        Contour id 2.0 has: 4.60523e-14 4.79728e-14  (19) (2 grids, 11079.0 11082.0)
        Contour id 3.0 has: 4.56287e-14 2.15439e-13  (26063) (32 grids, 11040.0 11103.0)
        Contour id 4.0 has: 4.56339e-14 7.18032e-14  (1669) (3 grids, 11039.0 11051.0)
"""

""" New with C extension:
yt.lagos   INFO       2008-04-11 00:18:03,943 Getting field Density from 197
yt.lagos   INFO       2008-04-11 00:18:05,109 Contouring over 1061337 cells with 30115 candidates
yt.lagos   INFO       2008-04-11 00:18:36,243 Getting field tempContours from 197
yt.lagos   INFO       2008-04-11 00:18:36,626 Identified 5 contours between 4.56276e-14 and 2.15510e-13
yt.lagos   INFO       2008-04-11 00:18:36,633 Getting field Contours from 197
yt.lagos   INFO       2008-04-11 00:18:37,601 Getting field GridIndices from 197
        Contour id 0.0 has: 3.43716e-14 9.88512e-14  (2352) (4 grids, 11081.0 11087.0)
        Contour id 1.0 has: 4.58381e-14 5.41707e-14  (13) (1 grids, 11079.0 11079.0)
        Contour id 2.0 has: 4.60523e-14 4.79728e-14  (19) (2 grids, 11079.0 11082.0)
        Contour id 3.0 has: 4.56287e-14 2.15439e-13  (26063) (32 grids, 11040.0 11103.0)
        Contour id 4.0 has: 4.56339e-14 7.18032e-14  (1669) (3 grids, 11039.0 11051.0)
"""

"""With C extension and no ->data[] accesses
yt.lagos   INFO       2008-04-11 14:19:24,079 Getting field Density from 197
yt.lagos   INFO       2008-04-11 14:19:25,356 Contouring over 1061337 cells with 30115 candidates
yt.lagos   INFO       2008-04-11 14:19:58,070 Getting field tempContours from 197
yt.lagos   INFO       2008-04-11 14:19:58,455 Identified 5 contours between 4.56276e-14 and 2.15510e-13
yt.lagos   INFO       2008-04-11 14:19:58,462 Getting field Contours from 197
yt.lagos   INFO       2008-04-11 14:19:59,418 Getting field GridIndices from 197
        Contour id 0.0 has: 4.56297e-14 9.88512e-14  (2351) (4 grids, 11081.0 11087.0)
        Contour id 1.0 has: 4.58381e-14 5.41707e-14  (13) (1 grids, 11079.0 11079.0)
        Contour id 2.0 has: 4.60523e-14 4.79728e-14  (19) (2 grids, 11079.0 11082.0)
        Contour id 3.0 has: 4.56287e-14 2.15439e-13  (26063) (32 grids, 11040.0 11103.0)
        Contour id 4.0 has: 4.56339e-14 7.18032e-14  (1669) (3 grids, 11039.0 11051.0)
"""


import sys
sys.path.insert(0,"/Users/matthewturk/Development/yt/trunk")
from yt.config import ytcfg

ytcfg["yt","LogLevel"] = '20'
ytcfg["yt","logFile"] = "False"
ytcfg["yt","suppressStreamLogging"] = "True"

con_field = "Temperature"
con_field = "Density"

import yt.lagos as lagos
import yt.raven as raven
import numpy as na
import scipy.weave as weave
import scipy.weave.converters as converters

print lagos, raven

#fn = "/Users/matthewturk/Research/data/yt_test_data/galaxy1360.dir/galaxy1360"
fn = "/Users/matthewturk/Research/data/DataDir0036/DataDump0036"
a = lagos.EnzoStaticOutput(fn)

v,c = a.h.find_max("Density")

sp = a.h.sphere(c,20000.0/a['au'])
sp2 = a.h.sphere(c,200.0/a['au'])

#lagos.identify_contours(sp, "Density", 1e-27, 1e-24)
#lagos.identify_contours(sp, "Density", 1e-33, 1e-22)
nc = 10
cons = na.logspace(na.log10(sp2[con_field].min()*0.9),
                   na.log10(sp2[con_field].max()),nc+1)
do_plot = 1
k = []
k_bad = []
for i in [8]:#range(nc):
    [grid.clear_data() for grid in sp._grids]
    mi, ma = cons[i], cons[i+1]
    print "Doing contour %s / %s (%0.5e %0.5e)" % (i+1,nc,mi,ma)
    mq=lagos.identify_contours(sp, con_field, mi, sp[con_field].max())
    for cid in mq:
        sp["Contours"][mq[cid]] = cid
    k.append(na.unique(sp["Contours"]).size)
    my_con = sp["Contours"]
    sp._flush_data_to_grids("Contours",-1)
    for cid in na.unique(sp["Contours"])[1:]:
        cid_ind = na.where(sp["Contours"] == cid)
        print "\tContour id %s has: %0.5e %0.5e  (%s) (%s grids, %s %s)" % \
            (cid, sp[con_field][cid_ind].min(), sp[con_field][cid_ind].max(),
             sp["Density"][cid_ind].size,
             na.unique(sp["GridIndices"][cid_ind]).size,
             na.unique(sp["GridIndices"][cid_ind]).min(),
             na.unique(sp["GridIndices"][cid_ind]).max(),
             )
    #sp = a.h.sphere(c,20000.0/a['au'])
    sp.get_data("Contours",in_grids=True)
    if not na.all(sp["Contours"] == my_con):
        khdoihiueh0
    g_bad = 0
    for grid in sp._grids:
        grid['dx'] = na.ones(grid.ActiveDimensions)*grid.dx
        grid['dy'] = na.ones(grid.ActiveDimensions)*grid.dy
        grid['dz'] = na.ones(grid.ActiveDimensions)*grid.dz
        g_bad+=lagos.check_neighbors(grid)
    print "GRID:",g_bad
    k_bad.append(lagos.check_neighbors(sp))
    print "SOURCE",k_bad[-1]
    if do_plot:
        #c=na.array([sp["x"][ind][-1],sp["y"][ind][-1],sp["z"][ind][-1]])
        pc = raven.PlotCollection(a, center=c)
        pc.add_slice("Density",0)
        pc.add_slice("Density",1)
        pc.add_slice("Density",2)
        for field in ["Density","Contours"]:
            pc.switch_field(field)
            pc.set_width(10000,'au')
            pc.save("test_bds_%02i_10000au" % i)
            pc.set_width(2000,'au')
            pc.save("test_bds_%02i_02000au" % i)
            pc.set_width(1000,'au')
            pc.save("test_bds_%02i_01000au" % i)
            pc.set_width(100,'au')
            pc.save("test_bds_%02i_00100au" % i)
    del sp.data["Contours"]
print k
print k_bad
